# -*- coding: utf-8 -*-
"""
Script with wrapped functions for profile generation.

Wrapping by using @jit decorators from numba library.

@author: Sergei Klykov
@licence: MIT, @year: 2024

"""
# %% Imports
import numpy as np
from typing import Union

# %% Checking and import the numba library for speeding up the calculation
global numba_installed
try:
    from numba import njit
    numba_installed = True
except ModuleNotFoundError:
    numba_installed = False


# %% Utility functions for round shaped objects (simple one for achieving acceleration by using numba library compilation)
@njit
def distance_f_acc(i_px: Union[int, float, np.ndarray], j_px: Union[int, float, np.ndarray], i_centre: Union[int, float],
                   j_centre: Union[int, float]) -> Union[float, np.ndarray]:
    """
    Calculate the distances for pixels accelerated by numba compilation.

    Parameters
    ----------
    i_px : int or numpy.ndarray
        Pixel(-s) i of an image.
    j_px : int or numpy.ndarray
        Pixel(-s) i of an image.
    i_centre : int | float
        Center of an image.
    j_centre : int | float
        Center of an image.

    Returns
    -------
    float or numpy.ndarray
        Distances between provided pixels and the center of an image.

    """
    return np.round(np.sqrt(np.power(i_px - i_centre, 2) + np.power(j_px - j_centre, 2)), 6)


@njit
def helper_calculations_bead(r: float, center_shifts: tuple) -> tuple:
    """
    Shift calculations to this function for compilation by numba.

    Parameters
    ----------
    r : float
        Radius.
    center_shifts : tuple
        Pixel shifts of the center.

    Returns
    -------
    tuple
        Couple of parameters required by the calling function.

    """
    max_size = int(round(2.5*r, 0))  # default size for the profile
    x_shift, y_shift = center_shifts  # unpacking the shift of the object center
    if abs(y_shift) > 0.0 or abs(x_shift) > 0.0:
        max_size += 1
        if abs(y_shift) >= 0.4 or abs(x_shift) >= 0.4:
            max_size += 1
    if max_size % 2 == 0:
        max_size += 1
    i_img_center = max_size // 2; j_img_center = max_size // 2
    if y_shift >= 0.0:
        i_center = i_img_center + y_shift
    else:
        i_center = i_img_center + y_shift + 1.0
    if x_shift >= 0.0:
        j_center = j_img_center + x_shift
    else:
        j_center = j_img_center + x_shift + 1.0
    net_shift = round(0.5*np.sqrt(y_shift*y_shift + x_shift*x_shift), 6)  # calculation the net shift of the picture center
    # Calculating the intensity distribution pixelwise with strict definition
    q_rad = round(0.25*r, 6)  # parameter for checking if a pixel is completely inside a circle
    size_subareas = 626  # number of subareas + 1 that can make defined number of steps, like np.linspace(0, 1, 11)
    single_point_value = 1.0/(size_subareas-1)  # single point non-zero value used below for finding number of points within circle border
    normalization = single_point_value*size_subareas*size_subareas  # normalization for summation of all points
    return max_size, i_center, j_center, q_rad, net_shift, size_subareas, single_point_value, normalization


# %% Round shape function (wrapper @njit and @jit don't work for the function below, most likely because of numpy.meshgrid function)
def discrete_shaped_bead_acc(r: float, center_shifts: tuple) -> np.ndarray:
    """
    Calculate the 2D shape of bead with the border pixel intensities defined from the counting area of these pixels within circle radius.

    Function accelerated by using compiled computation functions called inside.

    Parameters
    ----------
    r : float
        Radius of a bead.
    center_shifts : tuple
        Shifts on axis of the bead center.

    Returns
    -------
    img : numpy.ndarray
        2D normalized shape of the bead.

    """
    max_size, i_center, j_center, q_rad, net_shift, size_subareas, single_point_value, normalization = helper_calculations_bead(r, center_shifts)
    img = np.zeros(dtype=np.float32, shape=(max_size, max_size))  # creating by default float image, normalized to 1.0 as the max intensity
    for i in range(max_size):
        for j in range(max_size):
            distance = distance_f_acc(i, j, i_center, j_center)  # distance from the center to the pixel
            pixel_value = 0.0  # meaning the intensity in the pixel defined on the rules below
            if distance < q_rad:  # The pixel lays completely inside of circle border
                pixel_value = 1.0  # entire pixel lays inside the circle
            elif q_rad <= distance <= r + net_shift + 1.0:  # The pixel is intersecting with the circle border
                stop_checking = False  # flag for quitting these calculations if the pixel is proven to lay completely inside the circle
                # First, sort out the pixels that lay completely within the circle, but the distance is more than quarter of R:
                if i < i_center:
                    i_corner = i - 0.5
                else:
                    i_corner = i + 0.5
                if j < j_center:
                    j_corner = j - 0.5
                else:
                    j_corner = j + 0.5
                # Below - distance to the most distant point of the pixel
                distance_corner = distance_f_acc(i_corner, j_corner, i_center, j_center)
                if distance_corner <= r:
                    pixel_value = 1.0; stop_checking = True
                # So, the pixel's borders can potentially are intersected by the circle,
                # calculate the estimated intersection area for pixel intensity
                if not stop_checking:
                    i_m = i - 0.5; j_m = j - 0.5; i_p = i + 0.5; j_p = j + 0.5
                    x_row = np.linspace(start=i_m, stop=i_p, num=size_subareas); y_col = np.linspace(start=j_m, stop=j_p, num=size_subareas)
                    coords = np.meshgrid(x_row, y_col); distances = distance_f_acc(coords[0], coords[1], i_center, j_center)
                    circle_arc_area1 = np.where(distances <= r, single_point_value, 0.0)  # assigning the non-zero for intersected points
                    S1 = round(np.sum(circle_arc_area1)/normalization, 6)
                    if S1 > 1.0:
                        S1 = 1.0  # in the rare cases the integration sum can be > 1.0 due to the limited precision of numerical integration
                    pixel_value = S1  # assigning the found square of the area laying inside a circle
            img[i, j] = pixel_value  # assign the computed intensity to the stored 2D profile
    return img


# %% Utility functions for ellipse shaped objects (simple one for achieving acceleration by using numba library compilation)
@njit
def ellipse_equation_acc(i_px: Union[int, float, np.ndarray], j_px: Union[int, float, np.ndarray], i_centre: Union[int, float],
                         j_centre: Union[int, float], a: Union[int, float], b: Union[int, float], angle: Union[int, float]) -> float:
    """
    Calculate the ellipse equation ratio for the defining if the point (i_px, j_px) lays inside an ellipse accelerated by numba compilation.

    Parameters
    ----------
    i_px : Union[int, float, np.ndarray]
        Point (pixel) coordinate i.
    j_px : Union[int, float, np.ndarray]
        Point (pixel) coordinate j.
    i_centre : Union[int, float]
        i coordinate of the center of ellipse (center of mass).
    j_centre : Union[int, float]
        j coordinate of the center of ellipse (center of mass).
    a : Union[int, float]
        'a' ellipse axis in pixels.
    b : Union[int, float]
        'b' ellipse axis in pixels.
    angle : Union[int, float]
        angle between 'a' axis and 'X' axis.

    Reference
    ---------
    https://en.wikipedia.org/wiki/Ellipse

    Returns
    -------
    float
        Calculated equation (ratio) for the point.

    """
    angle = -angle  # makes the angle assignment consistent with the counter-clockwise count
    # Source for the equations below: https://en.wikipedia.org/wiki/Ellipse
    a_h2 = 0.25*a*a; b_h2 = 0.25*b*b  # half of axis (width and height) are used in the equations
    X = (j_px - j_centre)*np.cos(angle) + (i_px - i_centre)*np.sin(angle)  # projected X coordinate
    Y = -(j_px - j_centre)*np.sin(angle) + (i_px - i_centre)*np.cos(angle)  # projected Y coordinate
    i_diff2 = np.power(Y, 2); j_diff2 = np.power(X, 2)  # the general equation for ellipse
    return np.round(j_diff2/a_h2 + i_diff2/b_h2, 6)  # analytical equation for the shifted ellipse


# %% Ellipse shape definition accelerated by using the compiled function for defining belonging to the ellipse
def discrete_shaped_ellipse_acc(sizes: tuple, angle: float, center_shifts: tuple) -> np.ndarray:
    """
    Calculate ellipse shape project on the pixels.

    Function accelerated by using compiled computation functions called inside.

    Parameters
    ----------
    sizes : tuple
        Sizes a, b for an ellipse.
    angle : float
        Angle (counter-clockwise) between b and X axes.
    center_shifts : tuple
        Pixelwise shifts of an ellipse center.

    Returns
    -------
    img : numpy.ndarray
        2D normalized shape of the bead.

    """
    a, b = sizes  # from the definition of an ellipse: length of 2 axis
    max_d = max(a, b)  # for defining the largest and smallest axis of an ellipse
    max_size = int(round(1.25*max_d, 0))  # default size for the profile
    x_shift, y_shift = center_shifts  # unpacking the shift of the object center
    if abs(y_shift) > 0.0 or abs(x_shift) > 0.0:
        max_size += 1
        if abs(y_shift) > 0.4 or abs(x_shift) > 0.4:
            max_size += 1
    if max_size % 2 == 0:
        max_size += 1
    img = np.zeros(dtype=np.float32, shape=(max_size, max_size))  # creating by default float image, normalized to 1.0 as the max intensity
    i_img_center = max_size // 2; j_img_center = max_size // 2
    i_center = i_img_center + y_shift; j_center = j_img_center + x_shift
    # Calculating the intensity distribution pixelwise with strict definition
    size_subareas = 626  # number of subareas + 1 that can make defined number of steps, like np.linspace(0, 1, 11)
    single_point_value = 1.0/(size_subareas-1)  # single point non-zero value used below for finding number of points within circle border
    normalization = single_point_value*size_subareas*size_subareas  # normalization for summation of all points
    for i in range(max_size):
        for j in range(max_size):
            distance = ellipse_equation_acc(i, j, i_center, j_center, a, b, angle)  # equation for the ellipse testing
            pixel_value = 0.0  # meaning the intensity in the pixel defined on the rules below
            if 0.0 <= distance < 4.0:  # The pixel is intersecting with the ellipse border
                stop_checking = False  # flag for quitting these calculations if the pixel is proven to lay completely inside the circle
                # First, sort out the pixels that lay completely within the ellipse border, but the distance is more than quarter:
                if i <= i_center:  # shift to the left
                    i_corner = i - 0.5
                else:
                    i_corner = i + 0.5
                if j < j_center:  # shift to the bottom
                    j_corner = j - 0.5
                else:
                    j_corner = j + 0.5
                # Below - distance to the most distant point of the pixel
                distance_corner = ellipse_equation_acc(i_corner, j_corner, i_center, j_center, a, b, angle)
                if distance_corner < 0.5:  # empirical value for estimation that the pixel is entirely inside an ellipse
                    pixel_value = 1.0; stop_checking = True
                # So, the pixel's borders can potentially are intersected by the circle, calculate the estimated intersection area for pixel I
                if not stop_checking:
                    i_m = i - 0.5; j_m = j - 0.5; i_p = i + 0.5; j_p = j + 0.5
                    x_row = np.linspace(start=i_m, stop=i_p, num=size_subareas); y_col = np.linspace(start=j_m, stop=j_p, num=size_subareas)
                    coords = np.meshgrid(x_row, y_col); distances = ellipse_equation_acc(coords[0], coords[1], i_center, j_center, a, b, angle)
                    circle_arc_area1 = np.where(distances <= 1.0, single_point_value, 0.0)  # assigning the non-zero for intersected grid points
                    S1 = round(np.sum(circle_arc_area1)/normalization, 6)
                    if S1 > 1.0:
                        S1 = 1.0  # in the rare cases the integration sum can be > 1.0 due to the limited precision of numerical integration
                    pixel_value = S1  # assigning the found square of the area laying inside a circle
            img[i, j] = pixel_value  # assign the computed intensity to the stored 2D profile
    return img


# %% Default exports from this module
__all__ = ['discrete_shaped_bead_acc', 'discrete_shaped_ellipse_acc']



# -*- coding: utf-8 -*-
"""
Tests of the main module methods.

The pytest library available on: https://docs.pytest.org/en/latest/contents.html
For running collected here tests, it's enough to run the command "pytest" from the repository location in the command line.

@author: Sergei Klykov
@licence: MIT

"""
# %% Imports
import numpy as np

if __name__ != "__main__":
    from ..fluoscene import UscopeScene, FluorObj
    from ..utils.comp_funcs import get_radius_gaussian, get_ellipse_sizes


# %% Tests
def test_scene_initialization():
    try:
        UscopeScene(width=1, height=6)
        assert False, "Wrong initialization (UscopeScene(width=1, height=6)) not thrown the error"
    except ValueError:
        pass
    try:
        UscopeScene(width=12, height=3)
        assert False, "Wrong initialization (UscopeScene(width=12, height=3)) not thrown the error"
    except ValueError:
        pass
    try:
        UscopeScene(width=4.5, height=6)
        assert False, "Wrong initialization (UscopeScene(width=4.5, height=6)) not thrown the error"
    except TypeError:
        pass
    try:
        UscopeScene(width=5, height=4, image_type='none')
        assert False, "Wrong initialization (UscopeScene(width=2, height=2, img_type='none')) not thrown the error"
    except ValueError:
        pass
    scene = UscopeScene(width=51, height=100, image_type='uint16'); scene2 = UscopeScene(width=31, height=101, image_type=np.uint16)
    assert scene.max_pixel_value == scene2.max_pixel_value, 'Wrong initialization for image types np.uint16 and "uint16" '
    scene = UscopeScene(width=51, height=100, image_type='float'); scene2 = UscopeScene(width=31, height=101, image_type=np.float64)
    assert abs(scene.max_pixel_value - scene2.max_pixel_value) < 1E-6, 'Wrong initialization for image types np.uint16 and "uint16" '


def test_fluorobj_initialization():
    # False parameters provided for initialization - tests should fail
    try:
        FluorObj(typical_size=-0.4)
        assert False, "Wrong initialization (FluorObj(typical_size=-0.4)) not thrown the error"
    except ValueError:
        pass
    try:
        FluorObj(typical_size=5, center_shifts=1.0)
        assert False, "Wrong initialization (FluorObj(typical_size=5, center_shifts=1.0) not thrown the error"
    except TypeError:
        pass
    try:
        FluorObj(typical_size=(4.0, 4.0, 0.25), center_shifts=(0.0, 0.1))
        assert False, "Wrong initialization FluorObj(typical_size=(4.0, 4.0, 0.25), center_shifts=(0.0, 0.1) not thrown the error"
    except TypeError:
        pass
    try:
        FluorObj(typical_size=(1.49, 1.0, 2.18), center_shifts=(0.0, 0.1), shape_type='el')
        assert False, ("Wrong initialization FluorObj(typical_size=(1.49, 1.0, 2.18), center_shifts=(0.0, 0.1), shape_type='el') "
                       + " - not thrown the error for typical sizes out of range")
    except ValueError:
        pass
    # Checking initialization logic
    flobj = FluorObj(typical_size=4.75); flobj.get_shape(); flobj.crop_shape()
    assert flobj.profile.shape[0] > 4 and flobj.profile.shape[1] > 4, f"Profile sizes out of the expected range (5, 5): {flobj.profile.shape}"
    flobj = FluorObj(typical_size=(4.2, 5.1, 0.25*np.pi), shape_type='el'); flobj.get_shape(accelerated=False)
    flobj.get_casted_shape(max_pixel_value=1921, image_type='uint16'); flobj.crop_shape(); flobj.crop_shape()
    assert flobj.profile.shape == flobj.casted_profile.shape, "Profile and casted profile shapes aren't equal or double cropping causes errors"
    flobj = FluorObj(typical_size=2.0); flobj.get_shape(accelerated=True)
    assert flobj.profile.shape[1] == 3 and flobj.profile.shape[0] == 3, ("Profile sizes out of the expected range (3, 3): "
                                                                         + f"{flobj.profile.shape}")
    flobj = FluorObj(typical_size=(3.2, 4.2, -0.51*np.pi), shape_type='el'); flobj.get_shape(accelerated=True); flobj.crop_shape()
    flobj.get_casted_shape(max_pixel_value=100.0, image_type=np.float64)
    assert np.max(flobj.casted_profile) >= 100.0, f"Profile not casted properly, max value: {np.max(flobj.profile)}"


def test_objects_generation():
    # Not accelerated generation testing
    circles = UscopeScene.get_round_objects(mean_size=8, size_std=1.5, intensity_range=(202, 253), n_objects=5)
    scene = UscopeScene(width=55, height=42, image_type='uint16')
    placed_circles = scene.set_random_places(circles, overlapping=False, touching=False, only_within_scene=True)
    scene.put_objects_on(placed_circles, save_only_objects_inside=True)
    assert len(placed_circles) <= len(circles), "Number of placed circles more than number of generated circles"
    precise_objs = UscopeScene.get_random_objects(mean_size=(3.81, 2.36), size_std=(0.46, 0.22), shapes='mixed',
                                                  intensity_range=(180, 251), n_objects=4)
    scene2 = UscopeScene(width=64, height=52)
    placed_objs = scene2.set_random_places(precise_objs, overlapping=False, touching=False, only_within_scene=True)
    scene2.put_objects_on(placed_objs, save_only_objects_inside=True)
    assert len(placed_objs) <= len(precise_objs), "Number of placed objects more than number of generated 'precise' objects"
    # Testing for found bug in ver. 0.0.2 - getting wrong sizes for samples
    n_tested_objs = 79
    robjs2 = UscopeScene.get_round_objects(mean_size=12, size_std=8, intensity_range=(230, 254), n_objects=n_tested_objs)
    assert len(robjs2) == n_tested_objs, f"Round object generation not creating {n_tested_objs} objects as expected"
    # Acceleration generation testing
    accelerated_method_called = False
    try:
        import numba; numba_not_installed = False
        if numba is not None and not numba_not_installed:
            scene2.precompile_methods()  # verbose call of precompilation
            objs3 = scene2.get_objects_acc(mean_size=(2.5, 1.5), size_std=(1.0, 0.65), intensity_range=(240, 255),
                                           n_objects=5, shapes='ellipse')
            assert len(objs3) == 5, f"Number of generation objects by accelerated method is less than 3: {len(objs3)}"
            placed_objs3 = scene2.set_random_places(objs3, overlapping=False, touching=False, only_within_scene=True)
            assert len(placed_objs3) > 0, "Problem with placing common objects, no generated objects placed"
            accelerated_method_called = True
    except (ModuleNotFoundError, ImportError):
        numba_not_installed = True
    if numba_not_installed:
        if accelerated_method_called:
            assert False, "Accelerated method called wrongly (should be called because 'numba' not installed"


def test_other_methods():
    scene4 = UscopeScene(width=63, height=57); objs4 = None
    try:
        import numba
        if numba is not None:
            objs4 = scene4.get_objects_acc(mean_size=(3.75, 3.0), size_std=(0.25, 0.19), intensity_range=(195, 235),
                                           n_objects=4, shapes='mixed')
    except (ModuleNotFoundError, ImportError):
        objs4 = scene4.get_random_objects(mean_size=(3.75, 3.0), size_std=(0.25, 0.19), intensity_range=(195, 235),
                                          n_objects=4, shapes='mixed')
    if objs4 is not None:
        placed_objs = scene4.set_random_places(objs4, overlapping=False, touching=False, only_within_scene=True)
        scene4.put_objects_on(placed_objs, save_only_objects_inside=True)
        noisy_img = UscopeScene.noise2image(scene4.image)
        assert noisy_img.shape == scene4.image.shape, "Shapes of input and output images for noise2image() method not equal"
        assert np.max(noisy_img) != np.max(scene4.image), "Max pixel values for noisy and source image should be different"
        noisy_img2 = UscopeScene.noise2image(scene4.image, mean_g=2, sigma_g=21, gain_p=0.5)
        assert np.max(noisy_img2) < np.max(scene4.image), "Gain hasn't reduced max pixel value in process of adding noise to an image"
    else:
        assert False, "Something wrong with 'get_random_objects' method (check 'test_other_methods' function)"


def test_radiuses_generation():
    for i in range(250):
        r = get_radius_gaussian(r=1.5, r_std=1.0, mean_size=1.5, size_std=1.0)
        assert r >= 0.5, f"Generated r < 0.5: {round(r, 3)}"
        a, b, angle = get_ellipse_sizes(mean_size=(3.0, 2.0), size_std=(2.0, 1.0))
        r_min = min(a, b); r_max = max(a, b)
        assert r_min > 1.0 and r_max > 1.5, f"Sizes for ellipse {a, b} less smallest values"


def test_cast_images():
    scene = UscopeScene(width=267, height=232, image_type=np.uint16)
    objs = scene.get_round_objects(mean_size=12, size_std=2, intensity_range=(0, 4094), n_objects=14, image_type=scene.img_type)
    objs = scene.set_random_places(objs); scene.put_objects_on(objs); scene.add_noise(mean_g=200); scene.show_scene()
    img_neg_norm = UscopeScene.cast_image(scene.image)
    assert np.min(img_neg_norm) == -1.0 and np.max(img_neg_norm) == 1.0, "Image casting ('cast_image') to range [-1.0, 1.0] has a problem"
    img_int8 = UscopeScene.cast_image(scene.image, option='int8'); int8min = np.iinfo(np.int8).min; int8max = np.iinfo(np.int8).max
    assert np.min(img_int8) >= int8min and np.max(img_int8) == int8max, "Image casting ('cast_image') to int8 range [-127, 127] has a problem"
    img_int16 = UscopeScene.cast_image(scene.image, option='int16'); int16max = np.iinfo(np.int16).max; int16min = np.iinfo(np.int16).min
    assert np.min(img_int16) >= int16min and np.max(img_int16) == int16max, (f"Image casting ('cast_image') to int16 range [-{int16max}, "
                                                                             + f"{int16max}] has a problem")
